/*
Copyright (c) Members of the EGEE Collaboration. 2004.
See http://www.eu-egee.org/partners/ for details on the
copyright holders.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

// File: edg-wl-dagad-test.cpp
// Author: Francesco Giacomini <Francesco.Giacomini@cnaf.infn.it>
// Copyright (c) 2002 EU DataGrid.

// $Id: glite-wms-dagad-test.cpp,v 1.2.6.1 2010/04/07 09:50:34 mcecchi Exp $

#include <iterator>
#include <iostream>
#include <string>
#include <algorithm>
#include <memory>
#include <functional>
#include <boost/tuple/tuple.hpp>
#include <classad_distribution.h>

#include "requestad/DAGAd.h"
#include "requestad/DAGAdManipulation.h"

#include "glite/wmsutils/classads/classad_utils.h"

namespace jdl = glite::jdl;
namespace utilities = glite::wmsutils::classads;

namespace std {

std::ostream&
operator<<(std::ostream& os,
           jdl::DAGAd::node_value_type const& node)
{
  return os << node.first << ": " << &node.second;
}

std::ostream&
operator<<(std::ostream& os,
           jdl::DAGAd::dependency_value_type const& dependency)
{
  return os << "(" << *dependency.first << ", " << *dependency.second << ")";
}

}

namespace {

void print(jdl::DAGAd const& dagad)
{
  std::cout << "*** ClassAd:\n";
  std::cout << dagad.ad() << std::endl;

  jdl::DAGAd::node_iterator node_b;
  jdl::DAGAd::node_iterator node_e;
  boost::tie(node_b, node_e) = dagad.nodes();

  std::cout << dagad.num_nodes() << " nodes:\n";
  std::copy(
    node_b,
    node_e,
    std::ostream_iterator<jdl::DAGAd::node_value_type>(std::cout, "\n")

  );

  jdl::DAGAd::dependency_iterator dep_b;
  jdl::DAGAd::dependency_iterator dep_e;
  boost::tie(dep_b, dep_e) = dagad.dependencies();

  std::cout << dagad.num_dependencies() << " dependencies:\n";
  std::copy(
    dep_b,
    dep_e,
    std::ostream_iterator<jdl::DAGAd::dependency_value_type>(std::cout, "\n")
  );
}

std::string const test_dagad(
  "["
  "  type = \"dag\";"
  "  max_nodes_running = 10;"
  "  nodes = ["
  "    nodeA = ["
  "      node_type = \"edg-jdl\";"
  "      description = ["
  "      ]"
  "    ];"
  "    nodeB = ["
  "      node_type = \"edg-jdl\";"
  "      description = ["
  "      ]"
  "    ];"
  "    nodeC = ["
  "      node_type = \"edg-jdl\";"
  "      description = ["
  "      ]"
  "    ];"
  "    dependencies = {"
  "      { nodeA, nodeB },"
  "      { nodeA, nodeC }"
  "    }"
  "  ];"
  "  meta = 3"
  "]"
);

class DAGNodeNameIs: public std::unary_function<jdl::DAGAd::node_value_type, bool>
{
  std::string m_name;
public:
  DAGNodeNameIs(std::string const& name)
    : m_name(name)
  {}
  bool operator()(jdl::DAGAd::node_value_type const& v) const
  {
    return v.first == m_name;
  }

};

} // {anonymous}

int
main()
try {

  std::auto_ptr<classad::ClassAd> ad(utilities::parse_classad(test_dagad));

  jdl::DAGAd dagad(*ad);

  assert(jdl::set_edg_jobid(
           dagad,
           "https://ibm139.cnaf.infn.it:9000/rCvqHoLBEFKXJswGigYPyg"
         ));
  assert(jdl::remove_edg_jobid(dagad));
  assert(jdl::set_edg_jobid(
           dagad,
           "https://ibm139.cnaf.infn.it:9000/rCvqHoLBEFKXJswGigYPyg"
         ));

  jdl::DAGNodeInfo node_info("a_node.con", "condor");

  jdl::DAGNodeInfo new_node_info(node_info);
  new_node_info.retry_count(5);
  assert(!dagad.add_node("nodeB", node_info));
  assert(!dagad.replace_node("nodeE", new_node_info));
  assert(dagad.add_node("nodeE", node_info));
  assert(dagad.replace_node("nodeE", new_node_info));
  assert(dagad.add_dependency("nodeC", "nodeE"));
  assert(!dagad.add_dependency("nodeA", "nodeB"));
  assert(!dagad.remove_node("nodeB"));
  assert(dagad.remove_dependency("nodeA", "nodeB"));
  assert(dagad.remove_node("nodeB"));

  std::vector<std::string> is;
  is.push_back("one");
  is.push_back("two");
  assert(set_input_sandbox(dagad, is));
  std::vector<std::string> is_read = get_input_sandbox(dagad);
  assert(is == is_read);
  assert(parent_child(dagad, "nodeA", "nodeC"));
  assert(!parent_child(dagad, "nodeF", "nodeA"));

  print(dagad);

  jdl::DAGAd::node_iterator begin, end;
  boost::tie(begin, end) = dagad.nodes();
  for ( ; begin != end; ++begin) {
  }

  return 0;

  {
    // look for a node
    jdl::DAGAd::node_iterator b;
    jdl::DAGAd::node_iterator e;
    boost::tie(b, e) = dagad.nodes();
    jdl::DAGAd::node_iterator it = std::find_if(b, e, DAGNodeNameIs("nodeA"));
    assert(it != e);
    jdl::DAGNodeInfo info = it->second;
    std::cout << "nodeA description: " << *info.description_ad() << std::endl;
  }
  
  {
    // look for a node (alternative solution)
    jdl::DAGAd::node_iterator it = dagad.find("nodeA");
    jdl::DAGAd::node_iterator b;
    jdl::DAGAd::node_iterator e;
    boost::tie(b, e) = dagad.nodes();
    assert(it != e);
    jdl::DAGNodeInfo info = it->second;
    std::cout << "nodeA description: " << *info.description_ad() << std::endl;
  }

} catch (jdl::InvalidDAG&) {
  std::cerr << "Invalid DAGAd\n";
  return 1;
} catch (jdl::InvalidNodeInfo&) {
  std::cerr << "Invalid Node Info\n";
  return 2;
}
