/*
Copyright (c) Members of the EGEE Collaboration. 2004.
See http://www.eu-egee.org/partners/ for details on the
copyright holders.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/*
 * edg-wl-jdl-jsdl-test.cpp
 * Copyright (c) 2001 The European Datagrid Project - IST programme, all rights reserved.
 * Contributors are mentioned in the code where appropriate.
 */

#include "glite/wmsutils/exception/Exception.h"
#include "glite/jdl/adconverter.h"
#include "glite/jdl/collectionad.h"
#include "glite/jdl/JDLAttributes.h"
#include "glite/jdl/ExpDagAd.h"
#include "glite-wms-jdl-common.h"



// JSDL specific includes:

#include <string.h>
#include <libxml/xmlmemory.h>
#include <libxml/debugXML.h>
#include <libxml/HTMLtree.h>
#include <libxml/xmlIO.h>
#include <libxml/DOCBparser.h>
#include <libxml/xinclude.h>
#include <libxml/catalog.h>
#include <libxslt/xslt.h>
#include <libxslt/xsltInternals.h>
#include <libxslt/transform.h>
#include <libxslt/xsltutils.h>

#include <stdarg.h>  // va_start


/*
* JsdlException CLASS definition:
*/
class JsdlException: public glite::wmsutils::exception::Exception {
	public:
		/**Exception Constructor */
		JsdlException(const std::string file,int line, const std::string method,int code,
			const std::string error_message);
		/** Exception Destructor */
		virtual ~JsdlException() throw (){};
		/** Check whether parsing errors have been found (Static megthod)*/
		static bool hasErrors();
		/** */
		enum jsdl_error_code{
			NO_ERROR,
			XSL_ERROR,
			JSDL_ERROR
		};
		std::string getDebugInfo(){return stack_str;};
	private:
		std::string stack_str;
};

/*
* JSDL CLASS definition:
*/
class Jsdl {
	public:
		/** Constructor with Parameters
		* @param styleSheetFile the path of the stylesheet file to be loaded
		* @param jsdlFile the path of the JSDL file to be loaded
		*/
		Jsdl(const std::string& styleSheetFile, const std::string& jsdlFile);
		/** Copy Constructor
		* @param jsdl the source instance to be copied
		*/
		Jsdl(const Jsdl& jsdl);
		/** Operator "=" performs a deep copy of the instance
		* @param jsdl the source instance to be copied
		*/
		virtual void operator=(const Jsdl& jsdl) ;
		/** Default DCestructor */
		virtual ~Jsdl() throw(){};
		/** Set the StyleSheet to be used for parsing Jsdl document
		* @param styleSheetFile the path of the stylesheet file to be loaded*/
		void setXslFile(const std::string& styleSheetFile);
		/** Set the Jsdl to be used when applying XSL transformation
		* @param jsdlFile the path of the JSDL file to be loaded*/
		void setXmlFile(const std::string& jsdlFile);
		/** Retrieve the StyleSheet used for parsing Jsdl document
		* @return styleSheetFile the path of the loaded stylesheet*/
		std::string getXslFile(){return styleSheetFile;};
		/** Retrieve the Jsdl used when applied XSL transformation
		* @return jsdlFile the path of the loaded JSDL*/
		std::string setXmlFile(){return jsdlFile;};
		/** Apply the desired transformation
		* @return the resulting XSL transformed instance (as a string)
		*/
		std::string applyXSLT();
	private:
		void detach();
		xsltStylesheetPtr styleSheet;
		xmlDocPtr jsdl, parsedRes;
		std::string styleSheetFile, jsdlFile;
};


// ############     IMPLMENETATION    #######################

using namespace std ;
using namespace glite::jdl ;





// HANDLER Implementation
std::vector <std::string> jsdl_stack_exc;
void jsdlErrorHandler(void *ctx, const char *fmt, ...) {
		char *s;
		va_list ap;
		string ems;
		va_start(ap, fmt);
		while (*fmt){
			switch(*fmt++) {
				case 's':{ /* string */
					s= va_arg(ap, char *);
					string pass = string(s);
					jsdl_stack_exc.push_back(string(s));
				}
					break;
				default:
					// No significant Information
					break;
				}
		}
		va_end(ap);
}



/*
* JsdlException CLASS implementation:
*/
JsdlException::JsdlException(const std::string file,int line,
	const std::string method,int code,
	const std::string error_message):Exception(file, line, method, code,"JsdlException"){
	this->error_message = error_message;
	// Cycle Errors
	for (unsigned int i =0; i<jsdl_stack_exc.size(); i++){
		this->stack_str+=jsdl_stack_exc[i];
	}
	jsdl_stack_exc.resize(0);
}
bool JsdlException::hasErrors(){return jsdl_stack_exc.size();}

/*
* JSDL CLASS Implementation:
*/
// Default Constructor
Jsdl::Jsdl(const std::string& styleSheetFile, const std::string& jsdlFile){
	xmlSubstituteEntitiesDefault(1);
	xmlLoadExtDtdDefaultValue = 1;
	setXslFile(styleSheetFile);
	setXmlFile(jsdlFile);
}
// Copy Constructor
Jsdl::Jsdl(const Jsdl& jsdl){
	xmlSubstituteEntitiesDefault(1);
	xmlLoadExtDtdDefaultValue = 1;
	setXslFile(jsdl.styleSheetFile);
	setXmlFile(jsdl.jsdlFile);
}
void Jsdl::operator=(const Jsdl& jsdl){
	xmlSubstituteEntitiesDefault(1);
	xmlLoadExtDtdDefaultValue = 1;
	setXslFile(jsdl.styleSheetFile);
	setXmlFile(jsdl.jsdlFile);
}

void Jsdl::detach(){
	// Release Memory
	xsltFreeStylesheet(styleSheet);
	xmlFreeDoc(parsedRes);
	xmlFreeDoc(jsdl);
	xsltCleanupGlobals();
	xmlCleanupParser();
}
// setXslFile method
void Jsdl::setXslFile(const std::string& styleSheetFile){
	styleSheet = xsltParseStylesheetFile((const xmlChar *)styleSheetFile.c_str());
	xsltSetGenericErrorFunc(styleSheet,jsdlErrorHandler);
	xmlSetGenericErrorFunc (styleSheet,jsdlErrorHandler);
	this->styleSheetFile=styleSheetFile;
}
// setXmlFile method
void Jsdl::setXmlFile(const std::string& jsdlFile){
	jsdl = xmlParseFile(jsdlFile.c_str());
	if (JsdlException::hasErrors()){
		throw JsdlException (__FILE__,__LINE__,"Jsdl::setXmlFile",
			JsdlException::JSDL_ERROR,"JSDL parsing error for file: "+jsdlFile);
	}
	this->jsdlFile=jsdlFile;
}

// applyTransform method
std::string Jsdl::applyXSLT(){
	string result;
	const char **params =NULL;
	parsedRes=xsltApplyStylesheet(styleSheet, jsdl, params);
	if (JsdlException::hasErrors()){
		// Some Parsing Error
		throw JsdlException (__FILE__,__LINE__,"Jsdl::applyXSLT",
			JsdlException::XSL_ERROR,"XSL parsing error");
	}
	if (parsedRes){
		// PARSED SUCCESS
		xmlChar *doc_txt_ptr;
		int doc_txt_len;
		xsltSaveResultToString (&doc_txt_ptr, &doc_txt_len, parsedRes, styleSheet);
		result= string((const char *)doc_txt_ptr);
	}else {
		// UNABLE to apply Transformation
		throw JsdlException (__FILE__,__LINE__,"Jsdl::applyXSLT",
			JsdlException::XSL_ERROR,"StyleSheet Applying error");
	}
	// Release allocated memory before returning
	detach();
	return result;
}



// ############     TEST SUITE    #######################



// TEST static useful methods

string getStyleSheetFile(){
	string styleSheetFile = getJdlFile();
	if (styleSheetFile==""){
		// DEFAULT VALUE
		styleSheetFile="jsdl2jdl.xslt";
	}
	title("styleSheetFile loading: " +styleSheetFile);
	return styleSheetFile;
}
string getJsdlFile(){
	string jsdlFile=getArg();
	if (jsdlFile==""){
		// DEFAULT VALUE
		jsdlFile="jsdl.xml";
	}
	title("jsdlFile loading: "+jsdlFile);
	return jsdlFile;
}


// UNIT TEST
class JsdlTestFixture : public CppUnit::TestFixture {
	CPPUNIT_TEST_SUITE( JsdlTestFixture );
	CPPUNIT_TEST( testJobClient );
	CPPUNIT_TEST( testJobServer );
	CPPUNIT_TEST( testJSDLclass );
	CPPUNIT_TEST_SUITE_END();
public:
	Jsdl createJsdl(){
		Jsdl jsdl(getStyleSheetFile(), getJsdlFile());
		Jsdl jsdl2= jsdl;
		return jsdl2;
	}
	// Parses a jsdl and create the corresponging JobAd instance
	// Retrieve the submission string
	void testJobClient() {
		Jsdl jsdl= createJsdl();
		glite::jdl::JobAd jab(jsdl.applyXSLT());
		jobClientSide(jab);
	}
	// Parses a jsdl and create the corresponging JobAd instance
	// Passing it to a possible JobAd server
	void testJobServer() {
		Jsdl jsdl(getStyleSheetFile(), getJsdlFile());
		glite::jdl::JobAd jab(jsdl.applyXSLT());
		jobServerSide(jab);
	}
	// Test some unexpected results
	void testJSDLclass() {
		string styleSheetFile=getStyleSheetFile();
		string jsdlFile=getJsdlFile();
		glite::jdl::JobAd jab(    (Jsdl(styleSheetFile, jsdlFile)).applyXSLT()    );
		title("Created JDL from jsdl: " +jab.toLines() );
		jsdl = new Jsdl(styleSheetFile, jsdlFile) ;
		jsdl->applyXSLT();
		// Using Copy Constructor:
		Jsdl *jsdlBis = new Jsdl(*jsdl) ;
		delete jsdl;
		// Using "=" operator:
		Jsdl jsdlTer = *jsdlBis;
		jsdlBis->applyXSLT();
		delete jsdlBis;
		glite::jdl::JobAd jabBis(jsdlTer.applyXSLT());
		title("Used copy constructors of different kind, result is: "
			+ jabBis.toLines());
	}
private:
	Jsdl *jsdl;
};


// UNIT TEST END
int unitTest(){
	// CPP UNIT PART
	// Debug mode control:
	setWCDM(false);
	CppUnit::TestResult controller ;
	CppUnit::TestResultCollector result ;
	controller.addListener(&result);
	CppUnit::TestRunner runner ;
	//    ACTUAL CPP UNIT NAME
	runner.addTest (JsdlTestFixture::suite());
	runner.run (controller);
	CppUnit::TextOutputter outputter (&result , std::cerr);
	outputter.write();
	return result.wasSuccessful()?0:1;
}


int main(int argc,char *argv[]){
	try{
	if( (argc<2) || (strcmp(argv[1],"--help")==0)|| (strcmp(argv[1],"help")==0)){
		cout << "Usage : " << argv[0] << " <test number> [<StyleSheet file> <Jsdl File>]" << endl ;
		cout <<"0) Unit Test!!!"<<endl; cout <<" - - - " << endl ;
		cout <<"1) testJobClient"<<endl;
		cout <<"2) testJobServer"<<endl;
		cout <<"3) testJSDLclass"<<endl;
		return 0;
	}
	if (argc>2){setJdlFile(argv[2]);}
	if (argc>3){setArg(argv[3]);}
	JsdlTestFixture test;
	switch (boost::lexical_cast<int>(argv[1])){
		case 0 : unitTest(); break;
		case 1: test.testJobClient(); break;
		case 2: test.testJobServer(); break;
		case 3: test.testJSDLclass(); break;
		default: break;
	}
	return 0;
	}catch  (JsdlException &exc){
		cout << "\nJsdlException Operation not performed" <<flush <<endl ;
		cout <<"Stack...:  " << exc.printStackTrace()<<  endl ;
		cout <<"Exc Name:  " << exc.getExceptionName()<< endl;
		cout <<"What:      " << exc.what()<< endl;
		cout <<"DebugInfo: " << exc.getDebugInfo()<< endl;
	}catch  (glite::wmsutils::exception::Exception &exc){
		cout << "\nOperation not performed" <<flush <<endl ;
		cout <<"Stack:     " << exc.printStackTrace()<<  endl ;
		cout <<"Exc Name:  " << exc.getExceptionName()<< endl;
		cout <<"What:      " << exc.what()<< endl;
	}catch  (std::exception &exc){
		cout << "\nOperation not performed. std::exception found:\n" ;
		cout <<  exc.what() << endl ;
		cout << "\nUsage : " << argv[0] << " <test number> [<JDL file>]" << endl ;
	}catch  (...){
		cout << "\nOperation not performed. Unknown Exception found!\n" ;
	}
return 1 ;
};
